@@ -2,14 +2,15 @@
 Provides the LuminosoClient object, a wrapper for making
 properly-authenticated requests to the Luminoso REST API.
 """
-
-from .auth import LuminosoAuth
+from __future__ import unicode_literals
+from .auth import LuminosoAuth, TokenAuth
 from .constants import URL_BASE
 from .errors import (LuminosoError, LuminosoAuthError, LuminosoClientError,
     LuminosoServerError, LuminosoAPIError)
+from .compat import types_not_to_encode, urlparse
 from getpass import getpass
 import os
-import requests0 as requests
+import requests
 import logging
 import json
 import time
@@ -22,12 +23,12 @@ class LuminosoClient(object):
     A LuminosoClient is a thin wrapper around the REST API documented at
     https://api.luminoso.com/v4. As such, you interact with it by calling its
     methods that correspond to HTTP methods: `.get(url)`, `.post(url)`,
-    `.put(url)`, and `.delete(url)`.
+    `.put(url)`, `.patch(url)`, and `.delete(url)`.
 
     These URLs are relative to a 'base URL' for the LuminosoClient. For
     example, you can make requests for a specific account's projects
     by creating a LuminosoClient for
-    `https://api.luminoso.com/v4/projects/<accountname>`,
+    `https://api.luminoso.com/v4/projects/<account_id>`,
     or you can go deeper to create a client that makes requests for a
     specific project.
 
@@ -48,7 +49,7 @@ class LuminosoClient(object):
     method: when it gets a path starting with `/`, it will go back to the
     `root_url` instead of adding to the existing URL.
     """
-    def __init__(self, auth, url, proxies=None):
+    def __init__(self, auth, url):
         """
         Create a LuminosoClient given an existing auth object.
 
@@ -56,10 +57,6 @@ class LuminosoClient(object):
         the authentication for you.
         """
         self._auth = auth
-        self._session = requests.session()
-        self._session.auth = auth
-        if proxies is not None:
-            self._session.proxies = proxies
         self.url = ensure_trailing_slash(url)
         self.root_url = get_root_url(url)
 
@@ -67,8 +64,8 @@ class LuminosoClient(object):
         return '<LuminosoClient for %s>' % self.url
 
     @classmethod
-    def connect(cls, url=None, username=None, password=None,
-                proxies=None, auto_login=True):
+    def connect(cls, url=None, username=None, password=None, token=None,
+                token_file=None, token_auth=True):
         """
         Returns an object that makes requests to the API, authenticated
         with the provided username/password, at URLs beginning with `url`.
@@ -89,14 +86,6 @@ class LuminosoClient(object):
         which is convenient if it matches your Luminoso username:
 
             client = LuminosoClient.connect()
-
-        `proxies` is a dictionary from URL schemes (like 'http') to proxy
-        servers, in the same form used by the `requests` module.
-
-        By default, this object will automatically re-login if its
-        authentication times out, which happens after ten minutes of inactivity
-        or one hour, whichever comes first. If for security reasons you do not
-        want this to happen, set `auto_login` to False.
         """
         auto_account = False
         if url is None:
@@ -109,28 +98,92 @@ class LuminosoClient(object):
             url = URL_BASE + '/' + url.lstrip('/')
             root_url = URL_BASE
 
-        logger.info('collecting credentials')
-        username = username or os.environ['USER']
-        if password is None:
-            password = getpass('Password for %s: ' % username)
-
-        logger.info('creating LuminosoAuth object')
-        auth = LuminosoAuth(username, password, url=root_url, proxies=proxies,
-                            auto_login=auto_login)
-
-        client = cls(auth, url, proxies)
+        if token_auth:
+            auth = cls._get_token_auth(username, password, token, token_file,
+                                       root_url)
+        else:
+            auth = cls._get_nontoken_auth(username, password, token, root_url)
+        client = cls(auth, url)
         if auto_account:
             client = client.change_path('/projects/%s' %
                 client._get_default_account())
         return client
 
+    @staticmethod
+    def _get_token_auth(username, password, token, token_file, root_url):
+        logger.info('creating TokenAuth object')
+        if token is None and username is None:
+            # If no token or username was specified, check for a token saved
+            # in a local file.
+            token_file = token_file or get_token_filename()
+            try:
+                token_dict = json.load(open(token_file))
+                token = token_dict.get(urlparse(root_url).netloc)
+            except IOError:
+                logger.info('unable to read file %s; not using saved token')
+        if token is None:
+            if username is None:
+                username = os.environ['USER']
+            if password is None:
+                password = getpass('Password for %s: ' % username)
+            auth = TokenAuth.from_user_creds(
+                username, password, url=root_url)
+        else:
+            if username is not None:
+                logger.warn('ignoring "username" argument (using token)')
+            if password is not None:
+                logger.warn('ignoring "password" argument (using token)')
+            auth = TokenAuth(token)
+
+        return auth
+
+    @staticmethod
+    def _get_nontoken_auth(username, password, token, root_url):
+        if username is None:
+            username = os.environ['USER']
+        if password is None:
+            password = getpass('Password for %s: ' % username)
+
+        logger.info('creating LuminosoAuth object')
+        if token is not None:
+            logger.warn(
+                'ignoring "token" argument (using username and password)')
+        return LuminosoAuth(username, password, url=root_url)
+
+    def save_token(self, token_file=None):
+        """
+        Obtain the user's long-lived API token and save it in a local file.
+        If the user has no long-lived API token, one will be created.
+        Returns the token that was saved.
+        """
+        tokens = self._json_request('get', self.root_url + '/user/tokens/')
+        long_lived = [token['type'] == 'long_lived' for token in tokens]
+        if any(long_lived):
+            dic = tokens[long_lived.index(True)]
+        else:
+            # User doesn't have a long-lived token, so create one
+            dic = self._json_request('post', self.root_url + '/user/tokens/')
+        token = dic['token']
+        token_file = token_file or get_token_filename()
+        if os.path.exists(token_file):
+            saved_tokens = json.load(open(token_file))
+        else:
+            saved_tokens = {}
+        saved_tokens[urlparse(self.root_url).netloc] = token
+        directory, filename = os.path.split(token_file)
+        if directory and not os.path.exists(directory):
+            os.makedirs(directory)
+        with open(token_file, 'w') as f:
+            json.dump(saved_tokens, f)
+        return token
+
     def _request(self, req_type, url, **kwargs):
         """
         Make a request via the `requests` module. If the result has an HTTP
         error status, convert that to a Python exception.
         """
         logger.debug('%s %s' % (req_type, url))
-        func = getattr(self._session, req_type)
+        func = getattr(self._auth.session, req_type)
         result = func(url, **kwargs)
         try:
             result.raise_for_status()
@@ -160,7 +213,7 @@ class LuminosoClient(object):
         its contents. Otherwise, return the contents of the 'result' value.
         """
         response = self._request(req_type, url, **kwargs)
-        json_response = response.json
+        json_response = response.json()
         if not json_response:
             logger.error("Received response with no JSON: %s %s" %
                          (response, response.content))
@@ -198,9 +251,7 @@ class LuminosoClient(object):
         """
         params = jsonify_parameters(params)
         url = ensure_trailing_slash(self.url + path.lstrip('/'))
-        return self._json_request('post', url,
-            data=params,
-        )
+        return self._json_request('post', url, data=params)
 
     def put(self, path='', **params):
         """
@@ -215,9 +266,7 @@ class LuminosoClient(object):
         """
         params = jsonify_parameters(params)
         url = ensure_trailing_slash(self.url + path.lstrip('/'))
-        return self._json_request('put', url,
-            data=params,
-        )
+        return self._json_request('put', url, data=params)
 
     def patch(self, path='', **params):
         """
@@ -232,9 +281,7 @@ class LuminosoClient(object):
         """
         params = jsonify_parameters(params)
         url = ensure_trailing_slash(self.url + path.lstrip('/'))
-        return self._json_request('patch', url,
-            data=params,
-        )
+        return self._json_request('patch', url, data=params)
 
     def delete(self, path='', **params):
         """
@@ -247,9 +294,7 @@ class LuminosoClient(object):
         """
         params = jsonify_parameters(params)
         url = ensure_trailing_slash(self.url + path.lstrip('/'))
-        return self._json_request('delete', url,
-            params=params,
-        )
+        return self._json_request('delete', url, params=params)
 
     # Operations with a data payload
     def post_data(self, path, data, content_type, **params):
@@ -417,18 +462,19 @@ class LuminosoClient(object):
         if base_path is None:
             base_path = 'jobs/id'
         path = '%s%d' % (ensure_trailing_slash(base_path), job_id)
-        logger.info('waiting')
-        count = 0
+        start = time.time()
+        next_log = 0
         while True:
-            count += 1
-            if count % 10 == 0:
-                self.keepalive()
             response = self.get(path)
             if response['stop_time']:
                 if response['success']:
                     return response
                 else:
                     raise LuminosoError(response)
+            elapsed = time.time() - start
+            if elapsed > next_log:
+                logger.info('Still waiting (%d seconds elapsed).', next_log)
+                next_log += 120
             time.sleep(interval)
 
     def get_raw(self, path, **params):
@@ -440,6 +486,15 @@ class LuminosoClient(object):
         url = ensure_trailing_slash(self.url + path.lstrip('/'))
         return self._request('get', url, params=params).text
 
+
+def get_token_filename():
+    """
+    Return the default filename for storing API tokens.
+    """
+    return os.path.join(os.path.expanduser('~'),
+                        '.luminoso',
+                        'tokens.json')
+
 def get_root_url(url):
     """
     If we have to guess a root URL, assume it contains the scheme,
@@ -462,9 +517,8 @@ def jsonify_parameters(params):
     transmitted accurately. Other types of data need to be encoded into JSON.
     """
     result = {}
-    for param, value in params.iteritems():
-        if (isinstance(value, int) or isinstance(value, long)
-            or isinstance(value, basestring)):
+    for param, value in params.items():
+        if isinstance(value, types_not_to_encode):
             result[param] = value
         else:
             result[param] = json.dumps(value)
